
// RocketPort Infinity/Express device driver for Linux
//
// Copyright (C) 2004-2011 Comtrol, Inc.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

// Kernel Synchronization:
//
// This driver has 2 kernel control paths - syscall handlers (rp_xxx()
// calls into the driver -- possibly from user mode) and the timer
// bottom half tasklet (poll_xxx()).  This is a polled driver,
// interrupts are not used.
//
// Common data is contained in the CHANNEL_T struct, one for each
// serial port, normally accessed by pointer ChP.  The following
// variables are considered critical, being common to both control
// paths, changed by both and affecting program control flow. They are
// therefore protected by a spinlock ChP->slock in
// poll_transmit_serial(), rp_write() and rp_putchar().
//
//   -  xmit_flag
//   -  xmit_cnt
//   -  xmit_fifo_room

#include <linux/init.h>
#include <linux/moduleparam.h>
#include <linux/version.h>
#include <linux/wait.h>
#include "rocket2.h"

#define checkpoint() printk(KERN_ALERT "_DEBUG_ File[%s]\t Line[%d]\n", __FILE__, __LINE__)

#ifndef VERSION_CODE
#  define VERSION_CODE(vers,rel,seq) ( ((vers)<<16) | ((rel)<<8) | (seq) )
#endif

#if LINUX_VERSION_CODE < VERSION_CODE(2,6,4)    //  No version < 2.6.4 
#  error "This kernel is too old: not supported by this file"
#endif

#ifndef TTYLAYER_NEW
#if LINUX_VERSION_CODE >= VERSION_CODE(2,6,16)  //  Version < 2.6.12 must use old tty_flip.h
#define TTYLAYER_NEW
#endif
#endif

#undef CONFIG_DEVFS_FS

#define POLL_PERIOD HZ/100      //  Polling period .01 seconds (10ms)

#ifdef MODVERSIONS
#include <config/modversions.h>
#endif

#include <linux/module.h>
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/kernel.h>
#include <linux/slab.h>

#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/tty.h>
#include <linux/tty_driver.h>
#include <linux/tty_flip.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/pci.h>
#include <asm/uaccess.h>
#include <linux/spinlock.h>
#include <linux/interrupt.h>

#include "unity.h"

static struct tty_driver *rocket_driver;
static struct timer_list rocket_timer;
static struct rocket_version driver_version = {ROCKET_VERSION, ROCKET_DATE};

static atomic_t num_ports_open;
static int numberof_boards;

// Global array of configuration data, indexed by board number
static CONTROLLER_T uController[MAX_NUMBEROF_BOARDS];


// Line number is the /dev/ttyRPx number (x), the Minor number.  We
// assign them sequentially, starting at zero.

static unsigned int nextLineNumber;
static unsigned short totalPorts=0;
static char* portModes;
module_param(portModes,charp,0);

//Comtrol RocketPort 2 Model Info Array - Every device and its attributes
static modelInfo_t modelInfo[] = {
    //ID    Model String                                  Ports   2nd_ASIC  SMPTE
    {0x40, "RocketPort Infinity, Octa, RJ45, Selectable",   8,       0},
    {0x41, "RocketPort Infinity 32, External Interface",   32,       1},
    {0x42, "RocketPort Infinity 8, External Interface",     8,       0},
    {0x43, "RocketPort Infinity 16, External Interface",   16,       0},
    {0x44, "RocketPort Infinity 4, Quad, DB, Selectable",   4,       0},
    {0x45, "RocketPort Infinity 8, Octa, DB, Selectable",   8,       0},
    {0x46, "RocketPort Infinity 4, External Interface",     4,       0},
    {0x47, "RocketPort Infinity, 4, RJ45",                  4,       0},
    {0x4a, "RocketPort Infinity Plus, Quad",                4,       0},
    {0x4b, "RocketPort Infinity Plus, Octa",                8,       0},
    {0x4c, "RocketModem Infinity III, 8",                   8,       0},
    {0x4d, "RocketModem Infinity III, 4",                   4,       0},
    {0x4e, "RocketPort Infinity Plus, 2, 232",              2,       0},
    {0x4f, "RocketPort Infinity 2, SMPTE",                  2,       0},
    {0x50, "RocketPort Infinity Plus, 4, RJ45",             4,       0},
    {0x51, "RocketPort Infinity Plus, 8, RJ11",             8,       0},
    {0x52, "RocketPort Infinity Octa, SMPTE",               8,       0,       1},

    {0x60, "RocketPort Express, Octa, RJ45, Selectable",    8,       0},
    {0x61, "RocketPort Express 32, External Interface",    32,       1},
    {0x62, "RocketPort Express 8, External Interface",      8,       0},
    {0x63, "RocketPort Express 16, External Interface",    16,       0},
    {0x64, "RocketPort Express 4, Quad, DB, Selectable",    4,       0},
    {0x65, "RocketPort Express 8, Octa, DB, Selectable",    8,       0},
    {0x66, "RocketPort Express 4, External Interface",      4,       0},
    {0x67, "RocketPort Express, 4, RJ45",                   4,       0},
    {0x68, "RocketPort Express, 9, RJ11",                   8,       0},
    {0x72, "RocketPort Express Octa, SMPTE",                8,       0,       1},
};



static int rp_init(void);
static void rp_cleanup_module(void);
module_init(rp_init);
module_exit(rp_cleanup_module);

#ifdef MODULE_LICENSE
MODULE_LICENSE("Dual BSD/GPL");
#endif

#define AiopClock    44236800
#define AiopPrescale 0

// Calculate _twice_ the desired divisor, add 1, then divide by two so
// that we get the closest integer value to the "real" number rather
// than always truncating towards zero.
static int AiopBaudDivisor(int baudRate)
{
    return (((2L*AiopClock/(16L*(AiopPrescale+1)))/baudRate-2L)+1L)/2L;
}

static CHANNEL_T *GetChannelStruct(int line);
static void SetFlowControl(CHANNEL_T *ChP, int flowControl);
static int SetBaudRate(CHANNEL_T *ChP, int baudRate);
static int __init init_PCI(void);
static int __init register_PCI(int cardnum, struct pci_dev *dev);
static modelInfo_t *GetModelInfo(unsigned long model);
static int GetNumberOfModels(void);
static CHANNEL_T *GetChannelStruct(int line);
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,20)
static void configure_r_port(CHANNEL_T *ChP, struct termios *old_termios);
#else
static void configure_r_port(CHANNEL_T *ChP, struct ktermios *old_termios);
#endif
static int block_til_ready(struct tty_struct *tty, struct file *filp, CHANNEL_T *ChP);
static int rocket_paranoia_check(CHANNEL_T *ChP, const char *routine);

// Timer Polled Function Prototypes
static void poll_main(unsigned long dummy);
static void poll_receive_serial(CHANNEL_T *ChP);
static void poll_transmit_serial(CHANNEL_T *ChP);
static void poll_handle_interrupts(CHANNEL_T *ChP);

// TTY Layer Function Prototypes
static int rp_open(struct tty_struct *tty, struct file *filp);
static void rp_close(struct tty_struct *tty, struct file *filp);
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,10)
static int rp_write(struct tty_struct *tty, int from_user, const unsigned char *buf, int count);
#else           //implies kernel >= 2,6,10
static int rp_write(struct tty_struct *tty, const unsigned char *buf, int count);
#endif          // kernel < 2,6,10

#if LINUX_VERSION_CODE < VERSION_CODE(2,6,26)
static void rp_put_char(struct tty_struct *tty, unsigned char ch);
#else           //implies kernel >= 2,6,26
static int rp_put_char(struct tty_struct *tty, unsigned char ch);
#endif          // kernel < 2,6,26

static int rp_write_room(struct tty_struct *tty);
static int rp_chars_in_buffer(struct tty_struct *tty);
static void rp_flush_buffer(struct tty_struct *tty);
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,39)
static int rp_ioctl(struct tty_struct *tty, struct file *file, unsigned int cmd, unsigned long arg);
static int rp_tiocmget(struct tty_struct *tty, struct file *file);
static int rp_tiocmset(struct tty_struct *tty, struct file *file,  unsigned int set, unsigned int clear);
#else
static int rp_ioctl(struct tty_struct *tty, unsigned int cmd, unsigned long arg);
static int rp_tiocmget(struct tty_struct *tty);
static int rp_tiocmset(struct tty_struct *tty,  unsigned int set, unsigned int clear);
#endif
static void rp_throttle(struct tty_struct *tty);
static void rp_unthrottle(struct tty_struct *tty);
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,20)
static void rp_set_termios(struct tty_struct *tty, struct termios *old_termios);
#else           //implies kernel >= 2,6,20
static void rp_set_termios(struct tty_struct *tty, struct ktermios *old_termios);
#endif          //kernel < 2,6,20
static void rp_stop(struct tty_struct *tty);
static void rp_start(struct tty_struct *tty);
static void rp_hangup(struct tty_struct *tty);

#if LINUX_VERSION_CODE < VERSION_CODE(2,6,27)
static void rp_break(struct tty_struct *tty, int break_state);
#else           //implies kernel >= 2,6,27
static int rp_break(struct tty_struct *tty, int break_state);
#endif          //  kernel < 2.6.27

static void rp_send_xchar(struct tty_struct *tty, char ch);
static void rp_wait_until_sent(struct tty_struct *tty, int timeout);

static int get_serial_info(CHANNEL_T *ChP, struct serial_struct *retinfo);
static int set_serial_info(CHANNEL_T *ChP, struct serial_struct *usr_serial);
static int get_modem_info(CHANNEL_T *ChP, unsigned int *value);
static int set_modem_info(CHANNEL_T *ChP, unsigned int cmd, unsigned int *value);
static void EnableInterrupts(CHANNEL_T *ChP);
static void DisableInterrupts(CHANNEL_T *ChP);
static unsigned long uGetMsr(CHANNEL_T * ChP);


static struct tty_operations rocket_ops = {
    .open = rp_open,
    .close = rp_close,
    .write = rp_write,
    .put_char = rp_put_char,
    .write_room = rp_write_room,
    .chars_in_buffer = rp_chars_in_buffer,
    .flush_buffer = rp_flush_buffer,
    .ioctl = rp_ioctl,
    .throttle = rp_throttle,
    .unthrottle = rp_unthrottle,
    .set_termios = rp_set_termios,
    .stop = rp_stop,
    .start = rp_start,
    .hangup = rp_hangup,
    .break_ctl = rp_break,
    .send_xchar = rp_send_xchar,
    .wait_until_sent = rp_wait_until_sent,
    .tiocmget = rp_tiocmget,
    .tiocmset = rp_tiocmset
};


#define sCtlNumToCtlPtr(cardnum) &uController[cardnum]

struct rocket_config {
    int line;
    int flags;
    int closing_wait;
    int close_delay;
    int port;
    int reserved[32];
};

// Used in the RCKP_GET_PORTS IOCTL to report model info to application space.

struct rocket_ports {
    int tty_major;
    int callout_major;
    rocketModel_t rocketModel[8];
};

// obsolete custom get/set config ioctl handling routines that have
// been replaced by standard handlers for TIOCGSERIAL and TIOCSSERIAL

static int get_config(CHANNEL_T *info, struct rocket_config *retinfo)
{
    struct rocket_config tmp;

    if (!retinfo) return -EFAULT;
    memset(&tmp, 0, sizeof (tmp));
    tmp.line = info->portNumber;
    tmp.flags = info->flags;
    tmp.close_delay = info->close_delay;
    tmp.closing_wait = info->closing_wait;

    if (copy_to_user(retinfo, &tmp, sizeof (*retinfo))) return -EFAULT;
    return 0;
}

static int set_config(CHANNEL_T *info, struct rocket_config *new_info)
{
    struct rocket_config new_serial;

    if (copy_from_user(&new_serial, new_info, sizeof (new_serial)))
        return -EFAULT;

#ifdef CAP_SYS_ADMIN
    if (!capable(CAP_SYS_ADMIN))
#else
    if (!suser())
#endif
    {
        if ((new_serial.flags & ~ROCKET_USR_MASK) != (info->flags & ~ROCKET_USR_MASK))
            return -EPERM;
        info->flags = ((info->flags & ~ROCKET_USR_MASK) | (new_serial.flags & ROCKET_USR_MASK));
        configure_r_port(info, 0);
        return 0;
    }

    info->flags = ((info->flags & ~ROCKET_FLAGS) | (new_serial.flags & ROCKET_FLAGS));
    info->close_delay = new_serial.close_delay;
    info->closing_wait = new_serial.closing_wait;

    configure_r_port(info, 0);
    return 0;
}

// This function fills in a rocket_ports struct with information about
// what boards/ports are in the system.  This info is passed to user
// space.  See setinfinity.c where the info is used to create the
// /dev/ttyRPx ports.

static int get_ports(CHANNEL_T *info, struct rocket_ports *retports)
{
    struct rocket_ports tmp;
    int board;

    if (!retports)
        return -EFAULT;
    memset(&tmp, 0, sizeof (tmp));
    tmp.tty_major = rocket_driver->major;

    for (board = 0; board < 4; board++) {
        tmp.rocketModel[board].model = (sCtlNumToCtlPtr(board))->rocketModel.model;
        strcpy(tmp.rocketModel[board].modelString, (sCtlNumToCtlPtr(board))->rocketModel.modelString);
        tmp.rocketModel[board].numPorts = (sCtlNumToCtlPtr(board))->rocketModel.numPorts;
        tmp.rocketModel[board].startingPortNumber = (sCtlNumToCtlPtr(board))->rocketModel.startingPortNumber;
    }

    if (copy_to_user(retports, &tmp, sizeof (*retports)))
        return -EFAULT;
    return 0;
}

static int get_version(CHANNEL_T *info, struct rocket_version *retvers)
{
    if (copy_to_user(retvers, &driver_version, sizeof (*retvers)))
        return -EFAULT;
    return 0;
}

// The module "startup" routine; it's run when the module is loaded.

static int __init rp_init(void)
{
    int retval;

    printk(KERN_ALERT "RocketPort Infinity/Express device driver module, version %s, %s\n", ROCKET_VERSION, ROCKET_DATE);
    nextLineNumber = 0;

    // Set up the tty driver structure and then register this driver
    // with the tty layer.

    rocket_driver = alloc_tty_driver(MAX_NUMBEROF_PORTS);
    if (!rocket_driver) {
        printk(KERN_INFO "RocketPort: failed to allocate tty driver\n");
        return -ENOMEM;
    }

    rocket_driver->owner = THIS_MODULE;
    rocket_driver->name = "ttyRP";
    rocket_driver->driver_name = "Comtrol RocketPort";
    rocket_driver->major = RP2_MAJOR_NUMBER;
    rocket_driver->minor_start = 0;
    rocket_driver->num = MAX_NUMBEROF_PORTS;
    rocket_driver->type = TTY_DRIVER_TYPE_SERIAL;
    rocket_driver->subtype = SERIAL_TYPE_NORMAL;
    rocket_driver->init_termios = tty_std_termios;

#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,18))
    rocket_driver->devfs_name = "ttrp/%d";
#endif

#if LINUX_VERSION_CODE >= VERSION_CODE(2,6,20)
    rocket_driver->init_termios.c_ispeed = 9600;
    rocket_driver->init_termios.c_ospeed = 9600;
#endif

#if LINUX_VERSION_CODE < VERSION_CODE(2,6,26)
    rocket_driver->tiocmget = rp_tiocmget;
    rocket_driver->tiocmset = rp_tiocmset;
#endif

#if defined(TTY_DRIVER_NO_DEVFS)
    rocket_driver->flags |= TTY_DRIVER_NO_DEVFS;
#endif
#if defined(TTY_DRIVER_DYNAMIC_DEV)
    rocket_driver->flags |= TTY_DRIVER_DYNAMIC_DEV;
#endif

    tty_set_operations(rocket_driver, &rocket_ops);

    retval = tty_register_driver(rocket_driver);
    if (retval < 0) {
        put_tty_driver(rocket_driver);
        printk(KERN_INFO "RocketPort: couldn't install tty driver (error %d)\n", -retval);
        return -1;
    }

    DOPEN("Comtrol RocketPort Infinity/Express driver registered using major number %d\n", rocket_driver->major);

    memset(uController, 0, sizeof(uController));
    numberof_boards = init_PCI();

    if (numberof_boards == 0) {
        printk(KERN_INFO "RocketPort: no Infinity/Express cards found. Unloading driver.\n");
        if (tty_unregister_driver(rocket_driver))
            printk(KERN_INFO "RocketPort: error unregistering driver\n");
        put_tty_driver(rocket_driver);
        return -ENXIO;
    }

    init_timer(&rocket_timer);
    rocket_timer.function = poll_main;

    return 0;
}

#ifdef MODULE

static void rp_cleanup_module(void)
{
    int retval, i;

    DOPEN("Comtrol RocketPort Driver Exiting \n");
    del_timer_sync(&rocket_timer);
    for(i=0; i < totalPorts; i++)
        tty_unregister_device(rocket_driver, i);
    retval = tty_unregister_driver(rocket_driver);
    if (retval)
        printk(KERN_INFO "RocketPort: error %d unregistering driver\n", -retval);
    put_tty_driver(rocket_driver);
    printk(KERN_INFO "RocketPort: done unloading driver.\n");
}

#endif //MODULE

// Probes for RocketPort Infinity cards, inits them if found. Returns
// number of cards found

static int __init init_PCI(void)
{
    struct pci_dev *dev = NULL;
    int count = 0, i, match;

    while ( (dev = pci_get_device(PCI_VENDOR_ID_RP, PCI_ANY_ID, dev)) ) {
        //walk through table values
        for (i = 0, match=0; i < GetNumberOfModels(); i++) {
            if ( dev->device == modelInfo[i].model ) {
                register_PCI(count,dev);
                count++;
                break;//stop looking
            }
        }
    }
    return count;
}


// Called when a RP2 PCI card is found.  Retrieves and stores model
// information, init's ASIC serial port hardware.  Returns 1 if OK, 0
// if fail.

static int __init register_PCI(int cardnum, struct pci_dev *dev)
{
    modelInfo_t *modelInfo;
    CONTROLLER_T *CtlP;
    unsigned long length, start;
    int ch;
    char *portMode;

    modelInfo = GetModelInfo(dev->device);
    if (!modelInfo) {
        printk(KERN_INFO "Not a RocketPort card, ignoring... \n");
        return 0;
    }

    if (!dev || pci_enable_device(dev)) {
        printk(KERN_INFO "Unable to enable RocketPort card, ignoring... \n");
        return 0;
    }

    /*  Take the next controller struct, start filling it in.  */
    CtlP = &uController[cardnum];
    memset(CtlP, 0, sizeof(CONTROLLER_T));
    CtlP->controllerNumber = cardnum;
    CtlP->rocketModel.startingPortNumber = nextLineNumber;
    CtlP->rocketModel.model = modelInfo->model;
    CtlP->rocketModel.numPorts = modelInfo->numPorts;
    CtlP->numPorts = modelInfo->numPorts;
    CtlP->has_2nd_asic = modelInfo->has_2nd_asic;
    strcpy(CtlP->rocketModel.modelString, modelInfo->modelString);
    DOPEN("Initializing RPInfinity PCI card - %s  /dev/ttyRP%d - %ld \n", CtlP->rocketModel.modelString,
          CtlP->rocketModel.startingPortNumber,  CtlP->rocketModel.startingPortNumber + CtlP->rocketModel.numPorts -1);

    /*  Get the PCI memory address for region 0 */
    length = pci_resource_len(dev, 0);
    start =  pci_resource_start(dev, 0);
    if (!length) {
        printk(KERN_INFO "PCI Mem Region 0 length is zero, aborting card %d init \n", cardnum);
        return 0;
    }

    CtlP->pBasePciRegion0 = ioremap(start, length);
    DOPEN("Ioremapping PCI memory for cardnumber: %d  region: %d  start: 0x%lx  length: 0x%lx remapped ptr: 0x%p\n",
          cardnum, 0, start, length, CtlP->pBasePciRegion0);
    if (!CtlP->pBasePciRegion0) {
        printk("ERROR: ioremapping PCI memory for cardnumber: %d  region: %d  start: 0x%lx  length: 0x%lx \n",
               cardnum, 0, start, length);
        return 0;
    }

    // Get the PCI memory address for region 1
    length = pci_resource_len(dev, 1);
    start =  pci_resource_start(dev, 1);
    if (!length) {
        printk(KERN_INFO "PCI Mem Region 1 length is zero, aborting card %d init \n", cardnum);
        return 0;
    }

    CtlP->pBasePciMemory = ioremap(start, length);
    DOPEN("Ioremapping PCI memory for cardnumber: %d  region: %d  start: 0x%lx  length: 0x%lx remapped ptr: 0x%p\n",
          cardnum, 1, start, length, CtlP->pBasePciMemory);
    if (!CtlP->pBasePciMemory) {
        printk("ERROR: ioremapping PCI memory for cardnumber: %d  region: %d  start: 0x%lx  length: 0x%lx \n",
               cardnum, 1, start, length);
        return 0;
    }

    // FPGA testing requirement - set these 2 registers on init.  Else,
    // all PCI memory accesses to region 1 will return FF's.

    WRITEL(4, CtlP->pBasePciRegion0 + 0x110);
    WRITEL(0, CtlP->pBasePciRegion0 + 0x11C);

    uGlobalReset(CtlP);
    uSetClockPrescale(CtlP, AiopPrescale);
    uSetTDMClockRate(CtlP);

    printk(KERN_INFO "portModes: '%s'\n", portModes ? portModes : "not specified -- defaulting to RS232");

    // Initialize the channels, hardware and the channel structs
    for (ch = 0; ch < CtlP->numPorts; ch++) {
        CHANNEL_T *ChP = &CtlP->channel[ch];

        memset(ChP, 0, sizeof(CHANNEL_T));
        if (ch < 16) {
            ChP->pBaseRegister = CtlP->pBasePciMemory + (ch * 0x40);
            ChP->ucodeAddr     = CtlP->pBasePciMemory + UCODE_SRAM_CORE0 + (ch * 0x80);
        } else {
            ChP->pBaseRegister = CtlP->pBasePciMemory + 0x1000 + ((ch - 16) * 0x40);
            ChP->ucodeAddr     = CtlP->pBasePciMemory + 0x1000 + UCODE_SRAM_CORE0 + ((ch - 16) * 0x80);
        }

        ChP->channelNumber = ch;
        ChP->portNumber = ch + nextLineNumber;
        ChP->magic = RPORT_MAGIC;
        ChP->closing_wait = 30000;  // milliseconds
        ChP->close_delay = 500;     // milliseconds

        // initialize our wait queues
        init_waitqueue_head(&ChP->open_wait);
        init_waitqueue_head(&ChP->close_wait);

        if (portModes && strlen(portModes) >= (ChP->portNumber+1)*3)
            portMode=portModes+(ChP->portNumber*3);
        else
            portMode=0;

        if (portMode)
            printk(KERN_INFO "port%d mode: '%c%c%c'\n",ChP->portNumber,portMode[0],portMode[1],portMode[2]);

        // Set RocketMode flags
        if (modelInfo->is_smpte)
            ChP->flags = ROCKET_IS_SMPTE | ROCKET_MODE_RS422;  //for SMPTE cards, always set mode to RS422
        else if (portMode && !strncmp(portMode,"232",3))
            ChP->flags = ROCKET_MODE_RS232;
        else if (portMode && !strncmp(portMode,"422",3))
            ChP->flags = ROCKET_MODE_RS422;
        else if (portMode && !strncmp(portMode,"485",3))
            ChP->flags = ROCKET_MODE_RS485;
        else if (portMode && !strncmp(portMode,"486",3))
            ChP->flags = ROCKET_MODE_RS485FDS;
        else if (portMode && !strncmp(portMode,"487",3))
            ChP->flags = ROCKET_MODE_RS485FDM;
        else
            ChP->flags = ROCKET_MODE_RS232;

        ChP->monitorCdChange = 1;

        //  Init hardware to good defaults
        SetFlowControl(ChP, FLOW_CONTROL_NONE);
        uSetBaudDivisor(ChP, AiopBaudDivisor(9600));
        uSetData8(ChP);
        uDisParity(ChP);
        uSetStop1(ChP);
        uClrDtr(ChP);
        uClrRts(ChP);
        spin_lock_init(&ChP->slock);

#if LINUX_VERSION_CODE < VERSION_CODE(2,6,10)
        //printk(KERN_INFO "register_PCI: Initializing tmp_buf_sem\n");
        init_MUTEX(&ChP->tmp_buf_sem);
#endif

        if (dev)
            tty_register_device(rocket_driver, nextLineNumber+ChP->channelNumber, NULL);

        DOPEN("Port ttyRP%d  assigned register base address 0x%p and ucode address 0x%p \n",
              ChP->portNumber, ChP->pBaseRegister, ChP->ucodeAddr);
    }

    printk(KERN_INFO "Comtrol RocketPort PCI controller #%d found - %s.  Assigned /dev/ttyRP%d - %d \n",
           cardnum, CtlP->rocketModel.modelString, nextLineNumber, nextLineNumber + CtlP->numPorts - 1);

    nextLineNumber += CtlP->rocketModel.numPorts;
    totalPorts += CtlP->rocketModel.numPorts;
    return 1;
}


// The top level polling routine.  Repeats every 1/100 HZ (10ms).

static void poll_main(unsigned long dummy)
{
    CONTROLLER_T *CtlP;
    CHANNEL_T *ChP = NULL;
    int ctrl, ch;
    unsigned long next_time = jiffies + POLL_PERIOD;

    /*  Walk through all the boards (ctrl's) */
    for (ctrl = 0; ctrl < numberof_boards; ctrl++) {
        /*  Get a ptr to the board's control struct */
        CtlP = &uController[ctrl];
        if (CtlP->pBasePciMemory == 0)
            continue;

        for (ch = 0; ch < CtlP->numPorts; ch++) {
            ChP = &CtlP->channel[ch];

            if (ChP->flags & ASYNC_NORMAL_ACTIVE) {
                if(uRcvDataAvail(ChP))
                    poll_receive_serial(ChP);
                poll_handle_interrupts(ChP);
                if (ChP->xmit_flag)
                    poll_transmit_serial(ChP);
            } else if (ChP->blocked_open) {
                poll_handle_interrupts(ChP);
            }
        }
    }

    /* Reset the timer so we get called at the next clock tick (10ms).  */
    if (atomic_read(&num_ports_open))
        mod_timer(&rocket_timer, ((next_time > jiffies) ? next_time : jiffies));
}

// Called (from timer poll), checks for interrupts set, handling the event.

static void poll_handle_interrupts(CHANNEL_T *ChP)
{
    //this function can flood a syslog so be careful using straight printk() if at all

    if ((ChP->flags & ASYNC_INITIALIZED) == 0) {
        printk(KERN_INFO "RPInfinity: WARNING: poll_handle_interrupts ttyRP%d called with ChP->flags & NOT_INIT\n", ChP->portNumber);
        return;
    }
    if (!ChP->rptty) {
        printk(KERN_INFO "RPInfinity: WARNING: poll_handle_interrupts ttyRP%d called with ChP->rptty==NULL\n", ChP->portNumber);
        return;
    }

    DINTR("poll_handle_interrupts - interrupt on ttyRP%d \n", ChP->portNumber);

    /* CD change  */
    if (uIntCdChange(ChP) && (ChP->monitorCdChange)) {

        if (uCdActive(ChP)) {
            DINTR("ttyRP%d CD now on \n", ChP->portNumber);
            ChP->cd_status = 1;
        } else {
            DINTR("ttyRP%d CD now off \n", ChP->portNumber);
            tty_hangup(ChP->rptty);
            ChP->cd_status = 0;
        }
        wake_up_interruptible(&ChP->open_wait);
    }

    // The rest of the interrupt sources should not be enabled, but walk
    // through them, which will clear the flags if nothing else.

    if (uIntDsrChange(ChP))
        DINTR("RP2: DSR change on ttyRP%d...\n", ChP->portNumber);

    if (uIntCtsChange(ChP))
        DINTR("RP2: CTS change on ttyRP%d...\n", ChP->portNumber);

    if (uSpecialReceiveInt(ChP)) {
        uClrSpecialReceiveInt(ChP);
        DINTR("RP2: Special receive on ttyRP%d...\n", ChP->portNumber);
    }

    if (uXmitFifoEmptyInt(ChP))
        DINTR("RP2: XMIT FIFO Empty on ttyRP%d...\n", ChP->portNumber);

    if (uIntRiChange(ChP))
        DINTR("RP2: RI change on ttyRP%d...\n", ChP->portNumber);

    if (uAddrRcvdInt(ChP))
        DINTR("RP2: Addr Rcvd on ttyRP%d...\n", ChP->portNumber);

    if (uXmitShiftRegEmptyInt(ChP))
        DINTR("RP2: XMIT Shift Reg Empty on ttyRP%d...\n", ChP->portNumber);

    if (uXmitFifoTriggerInt(ChP))
        DINTR("RP2: XMIT FIFO Trigger on ttyRP%d...\n", ChP->portNumber);

    if (uInterCharTimeoutInt(ChP))
        DINTR("RP2: Interchar Timeout on ttyRP%d...\n", ChP->portNumber);
}


// Serial port receive data function.  Called (from timer poll) when
// an AIOPIC signals that receive data is present on a serial port.
// Pulls data from FIFO, moves it into the tty layer.

static void poll_receive_serial(CHANNEL_T *ChP)
{
    int rxFifoCnt, ttySpace=0, i;
    int numToRcv;                      /*  Number we will try to receive */
    int numRcvd = 0;                   /*  Actual number we will send up to the TTY layer */
    unsigned short byteStatus;

#ifdef TTYLAYER_NEW
    unsigned char *cbuf, *chead;
    char *fbuf, *fhead;
    struct tty_ldisc *ld;
    ld = tty_ldisc_ref(ChP->rptty);
#else
    unsigned char *cbuf;
    char *fbuf;
#endif

    if ((ChP->flags & ASYNC_INITIALIZED) == 0) {
        printk(KERN_INFO "RPInfinity: WARNING: poll_receive_serial ttyRP%d called with ChP->flags & NOT_INIT\n", ChP->portNumber);
        return;
    }
    if (!ChP->rptty) {
        printk(KERN_INFO "RPInfinity: WARNING: poll_receive_serial ttyRP%d called with ChP->rptty==NULL\n", ChP->portNumber);
        return;
    }

    // Determine how many we can read in.
    rxFifoCnt = uGetRxCnt(ChP);

#ifdef TTYLAYER_NEW
    ttySpace = ChP->rptty->receive_room;
#else
    ttySpace = ChP->rptty->ldisc.receive_room(ChP->rptty);
#endif

    if (ttySpace > 2 * TTY_FLIPBUF_SIZE)
        ttySpace = 2 * TTY_FLIPBUF_SIZE;

#ifndef TTYLAYER_NEW
    cbuf = ChP->rptty->flip.char_buf;
    fbuf = ChP->rptty->flip.flag_buf;
#endif

    if (rxFifoCnt > ttySpace)
        numToRcv = ttySpace;
    else
        numToRcv = rxFifoCnt;

#ifdef TTYLAYER_NEW
    numToRcv = tty_prepare_flip_string_flags(ChP->rptty, &chead, &fhead, numToRcv); // +
    if (numToRcv <= 0)
        goto done;
    cbuf = chead;
    fbuf = fhead;
#else
    if (numToRcv <= 0)
        return;
#endif

    DREAD("poll_receive_serial ttyRP%d - rxFifoCnt: %d  ttySpace: %d \n", ChP->portNumber,  rxFifoCnt, ttySpace);

    // If channel status indicates errors somewhere in the RX FIFO data,
    // then read both the status & data, else just data.

    if (uGetChanStatus(ChP) & ChanStatus_RxErr) {
        // there are errors, read data/status the "slow" way
        for (i = 0; i < numToRcv; i++) {
            byteStatus = uRxByteStatus(ChP);

            // The flag_buf corresponding to each char_buf can hold only
            // 1 error status.  Check for error conditions, set flag.
            //
            // Also, the TTY layer allows you to ignore
            // parity/framing/overrun/break errors.  We interpret that
            // to mean discard problematic bytes if the corresponding
            // ignore flag is set.  There is some disagreement in the
            // serial layer docs about whether to drop problematic bytes
            // or pass them up when IGNPAR is set.  We inherited the
            // drop action from Ted Ts'o.

            if (uRcvByteBreakDetected(byteStatus)) {
                if (ChP->ignore_status & U_BREAK)
                    continue;
                else if (ChP->report_status & U_BREAK)
                    *fbuf++ = TTY_BREAK;
            } else if (uRcvByteParityError(byteStatus)) {
                if (ChP->ignore_status & U_PARITY)
                    continue;
                else if (ChP->report_status & U_PARITY)
                    *fbuf++ = TTY_PARITY;
            } else if (uRcvByteFramingError(byteStatus)) {
                if (ChP->ignore_status & U_FRAMING)
                    continue;
                else if (ChP->report_status & U_FRAMING)
                    *fbuf++ = TTY_FRAME;
            } else if (uRcvByteRcvOverrun(byteStatus)) {
                if (ChP->ignore_status & U_OVERRUN)
                    continue;
                else if (ChP->report_status & U_OVERRUN)
                    *fbuf++ = TTY_OVERRUN;
            } else if(ChP->report_status)
                *fbuf++ = TTY_NORMAL;

            *cbuf++ = byteStatus & 0x00FF;
            numRcvd++;
        }
    } else {
        // no errors, read data in most efficient manner */
        unsigned int *dPtr = (unsigned int *)cbuf;
        int steps, remainder;

        steps = numToRcv / 4;
        remainder = numToRcv % 4;
        if (steps)
            for (i = 0; i < steps; i++)
                *dPtr++ = uRxDword(ChP);

        cbuf = (unsigned char *)dPtr;
        if (remainder)
            for (i = 0; i < remainder; i++) {
                byteStatus = uRxByteStatus(ChP);
                *cbuf++ = (unsigned char)(byteStatus & 0x00FF);
            }

        numRcvd = numToRcv;
        memset(fbuf, 0, numRcvd);
    }

    // Push the data up to the tty layer
#ifdef TTYLAYER_NEW
    tty_flip_buffer_push(ChP->rptty);
#else
    ChP->rptty->ldisc.receive_buf(ChP->rptty, ChP->rptty->flip.char_buf, ChP->rptty->flip.flag_buf, numRcvd);
#endif

#ifdef TTYLAYER_NEW
done:
    tty_ldisc_deref(ld);
#endif
}

// Serial port transmit data function.  Called from the timer polling
// loop as a result of xmit_flag being set in the channel struct,
// indicating data (from the tty layer) is ready to be sent out the
// serial port.  Data is buffered in channel xmit_buf, it is moved to
// the port's xmit FIFO.

static void poll_transmit_serial(CHANNEL_T *ChP)
{
    int numToXmit;                     /*  Number we will transmit */

    if ((ChP->flags & ASYNC_INITIALIZED) == 0) {
        printk(KERN_INFO "RPInfinity: WARNING: poll_transmit_serial ttyRP%d called with ChP->flags & NOT_INIT\n", ChP->portNumber);
        return;
    }
    if (!ChP->rptty) {
        printk(KERN_INFO "RPInfinity: WARNING: poll_transmit_serial ttyRP%d called with tty==NULL\n", ChP->portNumber);
        return;
    }

    spin_lock_bh(&ChP->slock);

    ChP->xmit_fifo_room = FIFO_SIZE - uGetTxCnt(ChP);

    spin_unlock_bh(&ChP->slock);
    DREAD("poll_transmit_serial ttyRP%d - xmit_fifo_room: %d \n", ChP->portNumber, ChP->xmit_fifo_room);

    while (1) {
        unsigned int *dPtr;
        unsigned char *bPtr;
        int i, remainder, steps;

        if (ChP->rptty->stopped || ChP->rptty->hw_stopped)
            break;

        numToXmit = MIN(ChP->xmit_fifo_room, MIN(ChP->xmit_cnt, XMIT_BUF_SIZE - ChP->xmit_tail));

        dPtr = (unsigned int *)(ChP->xmit_buf + ChP->xmit_tail);
        steps = numToXmit / 4;
        remainder = numToXmit % 4;

        if (steps)
            for (i = 0; i < steps; i++)
                uXmitDword(ChP, *dPtr++);

        if (remainder) {
            bPtr = (unsigned char *)dPtr;
            for (i = 0; i < remainder; i++)
                uXmitByte(ChP, *bPtr++);
        }

        spin_lock_bh(&ChP->slock);
        ChP->xmit_tail += numToXmit;
        ChP->xmit_tail &= XMIT_BUF_SIZE - 1;
        ChP->xmit_cnt -= numToXmit;
        ChP->xmit_fifo_room -= numToXmit;
        DWRITE(KERN_INFO "poll_transmit_serial (ttyRP%d) number transmitted: %d \n", ChP->portNumber, numToXmit);

        if (ChP->xmit_cnt == 0)
            ChP->xmit_flag = 0;

        spin_unlock_bh(&ChP->slock);

        if (ChP->xmit_cnt < WAKEUP_CHARS) {
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,10)
            if ((ChP->rptty->flags & (1 << TTY_DO_WRITE_WAKEUP)) && ChP->rptty->ldisc.write_wakeup)
                (ChP->rptty->ldisc.write_wakeup) (ChP->rptty);
            wake_up_interruptible(&ChP->rptty->write_wait);
#else
            tty_wakeup(ChP->rptty);
#endif
        }

        DWRITE("(%d,%d,%d,%d)...", ChP->xmit_cnt, ChP->xmit_head, ChP->xmit_tail, ChP->xmit_fifo_room);//uncommented 3-2 BKA

        if (numToXmit <= 0)
            break;
    }
}



//  handler that opens a serial port.  Creates xmit_buf storage, fills in
//  port's r_port struct.  Initializes the port hardware.

static int rp_open(struct tty_struct *tty, struct file *filp)
{
    int line;   // The port minor number ('x' in ttyRPx)
    int ret;
    CHANNEL_T *ChP;
    unsigned long page;

//compiler complains if this isn't conditional
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,10)
    unsigned long paget;
#endif          //kernel < 2,6,20

    line = TTY_GET_LINE(tty);
    //printk("DEBUG rp_open: line = %d\n", line);

    if ((line < 0) || (line >= MAX_NUMBEROF_PORTS)) {
        printk(KERN_INFO "RPInfinity:  Error opening port %d - Invalid port number \n", line);
        return -ENXIO;
    }

    ChP = GetChannelStruct(line);
    if (!ChP) {
        printk(KERN_INFO "RPInfinity:  Error opening port %d - No channel struct found \n", line);
        return -ENXIO;
    }

    DOPEN("rp_open ttyRP%d \n", ChP->portNumber);

    page = __get_free_page(GFP_KERNEL);
    if (!page) {
        printk(KERN_INFO "RPInfinity: ttyRP%d Error opening port - Unable to retrieve mem page \n", ChP->portNumber);
        return -ENOMEM;
    }

#if LINUX_VERSION_CODE < VERSION_CODE(2,6,10)
    paget = __get_free_page(GFP_KERNEL);
    if (!paget) {
        free_page(page);
        return -ENOMEM;
    }
#endif          //kernel < 2,6,10

    if (ChP->flags & ASYNC_CLOSING) {
        wait_event_interruptible(ChP->close_wait,(!(ChP->flags & ASYNC_CLOSING)));
        free_page(page);
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,10)
        free_page(paget);
#endif          //kernel < 2,6,10
        return ((ChP->flags & ASYNC_HUP_NOTIFY) ? -EAGAIN : -ERESTARTSYS);
    }

    /*  We must not sleep from here until the port is marked fully in use.  */
    if (ChP->xmit_buf)
        free_page(page);
    else
        ChP->xmit_buf = (unsigned char *) page;

#if LINUX_VERSION_CODE < VERSION_CODE(2,6,10)
    if (ChP->tmp_buf)
        free_page(paget);
    else
        ChP->tmp_buf = (unsigned char *)paget;
#endif          //kernel < 2,6,10

    tty->driver_data = ChP;
    ChP->rptty = tty;

    if (ChP->count++ == 0) {
        atomic_inc(&num_ports_open);
    }

    //  ChP->count is now 1; so its safe to sleep now.

    if (!(ChP->flags & ASYNC_INITIALIZED)) {
        uChannelReset(ChP);
        if (uCdActive(ChP))
            ChP->cd_status = 1;
        else
            ChP->cd_status = 0;
        uFlushRxFifo(ChP);
        uFlushTxFifo(ChP);
        EnableInterrupts(ChP);

        uEnRxFIFO(ChP);

        uSetXon(ChP);
        uLoadUcode(ChP);
        uStartRxProcessor(ChP);
        uEnTransmit(ChP);

        ChP->flags |= ASYNC_INITIALIZED;

        spin_lock_bh(&ChP->slock);
        ChP->xmit_flag = 0;
        spin_unlock_bh(&ChP->slock);

        //Set up the tty->alt_speed kludge
        if ((ChP->flags & ASYNC_SPD_MASK) == ASYNC_SPD_HI)
            ChP->rptty->alt_speed = 57600;
        if ((ChP->flags & ASYNC_SPD_MASK) == ASYNC_SPD_VHI)
            ChP->rptty->alt_speed = 115200;
        if ((ChP->flags & ASYNC_SPD_MASK) == ASYNC_SPD_SHI)
            ChP->rptty->alt_speed = 230400;
        if ((ChP->flags & ASYNC_SPD_MASK) == ASYNC_SPD_WARP)
            ChP->rptty->alt_speed = 460800;
        if ((ChP->flags & ASYNC_SPD_MASK) == ASYNC_SPD_WARP13)
            ChP->rptty->alt_speed = 921600;

        configure_r_port(ChP, NULL);

        if (tty->termios->c_cflag & CBAUD) {
            uSetDtr(ChP);
            uSetRts(ChP);
        }
    }

    /*  Starts (or resets) the main polling loop */
    mod_timer(&rocket_timer, jiffies + POLL_PERIOD);

    ret = block_til_ready(tty, filp, ChP);
    if (ret) {
        DOPEN("rp_open returning after block_til_ready with %d\n", ret);
        return ret;
    }
    DOPEN("rp_open ttyRP%d complete, count=%d \n", ChP->portNumber, ChP->count);

    return 0;
}

/*
 *  Configures a rocketport port according to its termio settings.  Called from
 *  user mode into the driver (syscall handler).  If old_termios is passed in,
 *  the original baud is retained if an invalid one is attempted.
 *  Inputs:  Channel pointer, and old termios (NULL OK).
 */
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,20)
static void configure_r_port(CHANNEL_T *ChP, struct termios *old_termios)
#else           //implies kernel >= 2,6,20
static void configure_r_port(CHANNEL_T *ChP, struct ktermios *old_termios)
#endif          //kernel <2,6,20
{
    unsigned long cflag;
    int bits = 0, baud;

    if (!ChP->rptty || !ChP->rptty->termios) {
        printk(KERN_INFO "RPInfinity:  configure_r_port - no TTY defined, exiting \n");
        return;
    }
    DOPEN("configure_port ttyRP%d \n", ChP->portNumber);
    cflag = ChP->rptty->termios->c_cflag;

    /* Byte size and parity */

    switch (cflag & CSIZE) {
    case CS5:
        uSetData5(ChP);
        bits = 7;
        break;

    case CS6:
        uSetData6(ChP);
        bits = 8;
        break;

    case CS7:
        uSetData7(ChP);
        bits = 9;
        break;

    case CS8:
    default:
        uSetData8(ChP);
        bits = 10;
        break;
    }

    if (cflag & CSTOPB) {
        uSetStop2(ChP);
        bits++;
    } else
        uSetStop1(ChP);

    if (cflag & PARENB) {
        uEnParity(ChP);
        bits++;
        if (cflag & CMSPAR) {
            if (cflag & PARODD)
                uSetMarkParity(ChP);
            else
                uSetSpaceParity(ChP);
        } else {
            if (cflag & PARODD)
                uSetOddParity(ChP);
            else
                uSetEvenParity(ChP);
        }
    } else {
        uDisParity(ChP);
    }

    baud = tty_get_baud_rate(ChP->rptty);

    if (baud == 38400 && (ChP->flags & ASYNC_SPD_MASK) == ASYNC_SPD_CUST) {
        if (ChP->baud_base == 0)
            baud = 38400;
        else if (ChP->custom_divisor)
            baud = ChP->baud_base / ChP->custom_divisor;
        else
            baud = ChP->baud_base;
    }

    if (baud) {
        SetBaudRate(ChP, baud);
        ChP->cps = baud / bits;
    }

    if (!ChP->cps) {
        DOPEN("configure_r_port ttyRP%d ERROR - cps is zero \n", ChP->portNumber);
        ChP->cps = 11520;
    }

    /*  Set flow control */
    if (cflag & CRTSCTS) {
        SetFlowControl(ChP, FLOW_CONTROL_RTSCTS);
    }
#if 0
    else if (I_IXON(ChP->rptty)) {
        SetFlowControl(ChP, FLOW_CONTROL_XONXOFF);
        if (I_IXANY(ChP->rptty))
            uEnXany(ChP);
        else
            uDisXany(ChP);
        uSetTxXONChar(ChP, START_CHAR(ChP->rptty));
        uSetTxXOFFChar(ChP, STOP_CHAR(ChP->rptty));
    }
#endif
    else {
        SetFlowControl(ChP, FLOW_CONTROL_NONE);
        uDisXany(ChP);
        uSetXon(ChP);
    }

    if (cflag & CLOCAL) {
        ChP->monitorCdChange = 0;
    } else {
        if (uCdActive(ChP))
            ChP->cd_status = 1;
        else
            ChP->cd_status = 0;
        ChP->monitorCdChange = 1;
    }

    /*
     *   TTY layer flags concerning whether to ignore (drop) bytes with errors,
     *   or whether to report them to the TTY layer.
     */
    ChP->report_status = U_OVERRUN;
    if (I_INPCK(ChP->rptty))
        ChP->report_status |= U_FRAMING | U_PARITY;
    if (I_BRKINT(ChP->rptty) || I_PARMRK(ChP->rptty))
        ChP->report_status |= U_BREAK;

    ChP->ignore_status = 0;
    if (I_IGNPAR(ChP->rptty))
        ChP->ignore_status |= U_FRAMING | U_PARITY;
    if (I_IGNBRK(ChP->rptty)) {
        ChP->ignore_status |= U_BREAK;
        /*
         * If we're ignoring parity and break indicators,
         * ignore overruns too.  (For real raw support).
         */
        if (I_IGNPAR(ChP->rptty))
            ChP->ignore_status |= U_OVERRUN;
    }

    if ((ChP->flags & ROCKET_RTS_TOGGLE) ||
            ((ChP->flags & ROCKET_MODE_MASK) == ROCKET_MODE_RS485) ||
            ((ChP->flags & ROCKET_MODE_MASK) == ROCKET_MODE_RS485FDS)) {
        uEnRTSToggle(ChP);
        uRtsTogglePolarityHigh(ChP);
    } else {
        uDisRTSToggle(ChP);
        uRtsTogglePolarityLow(ChP);
    }

    // Set mode to RS422 for SMPTE cards
    if ((ChP->flags & ROCKET_IS_SMPTE) ||
            ((ChP->flags & ROCKET_MODE_MASK) == ROCKET_MODE_RS422)) {
        DOPEN("configure_r_port ttyRP%d Interface Set To RS422\n", ChP->portNumber);
        uSetInterfaceMode(ChP, URS_RS422);
        SetFlowControl(ChP, FLOW_CONTROL_NONE);
    } else if ((ChP->flags & ROCKET_MODE_MASK) == ROCKET_MODE_RS485) {
        DOPEN("configure_r_port ttyRP%d Interface Set To RS485 2-wire\n", ChP->portNumber);
        uSetInterfaceMode(ChP, URS_RS485_HD);
        SetFlowControl(ChP, FLOW_CONTROL_NONE);
    } else if ((ChP->flags & ROCKET_MODE_MASK) == ROCKET_MODE_RS485FDS) {
        DOPEN("configure_r_port ttyRP%d Interface Set To RS485 4-wire Slave\n", ChP->portNumber);
        uSetInterfaceMode(ChP, URS_RS485_FD_SLAVE);
        SetFlowControl(ChP, FLOW_CONTROL_NONE);
    } else if ((ChP->flags & ROCKET_MODE_MASK) == ROCKET_MODE_RS485FDM) {
        DOPEN("configure_r_port ttyRP%d Interface Set To RS485 4-wire Master\n", ChP->portNumber);
        uSetInterfaceMode(ChP, URS_RS485_FD_MASTER);
        SetFlowControl(ChP, FLOW_CONTROL_NONE);
    } else {
        DOPEN("configure_r_port ttyRP%d Interface Set To RS232\n", ChP->portNumber);
        uSetInterfaceMode(ChP, URS_RS232);
    }

}

// syscall handler that closes a serial port.

static void rp_close(struct tty_struct *tty, struct file *filp)
{
    int timeout;
    CHANNEL_T *ChP = (CHANNEL_T *)tty->driver_data;

    if (rocket_paranoia_check(ChP, "rp_close"))
        return;
    DCLOSE("rp_close ttyRP%d, count = %d\n", ChP->portNumber, ChP->count);

    if (tty_hung_up_p(filp)) {
        printk(KERN_INFO "rp_close ttyRP%d tty hung up  \n", ChP->portNumber);
        return;
    }

    if (--ChP->count < 0) {
        printk(KERN_INFO "rp_close: bad serial port count for ttyRP%d: %d\n", ChP->portNumber, ChP->count);
        ChP->count = 0;
    }
    if (ChP->count) {
        DCLOSE("rp_close ttyRP%d count (%d) non-zero after decrement -- not shutting down port.\n", ChP->portNumber, ChP->count);
        return;
    }

    ChP->flags |= ASYNC_CLOSING;

    /*  Notify the line discpline to only process XON/XOFF characters  */
    tty->closing = 1;

    /* If transmission was throttled by the application request, just flush the xmit buffer. */
    if (tty->flow_stopped)
        rp_flush_buffer(tty);

    /* Wait for the transmit buffer to clear  */
    if (ChP->closing_wait != ASYNC_CLOSING_WAIT_NONE) {
        DCLOSE("rp_close ttyRP%d - waiting for transmit buffer to clear \n", ChP->portNumber);
        tty_wait_until_sent(tty, msecs_to_jiffies(ChP->closing_wait));
    }


    //Before we drop DTR, make sure the UART transmitter has completely drained.
    //**and don't call a wait if we don't have to.**
    if(uGetTxCnt(ChP) > 0) {
        if(ChP->cps > 0)
            timeout = (uGetTxCnt(ChP) + 1) * HZ / ChP->cps;
        else
            timeout = 1;

        DCLOSE("rp_close ttyRP%d - waiting for FIFO to drain, or timeout (%d)  \n", ChP->portNumber, timeout);

        rp_wait_until_sent(tty, timeout);

    } else {
        DCLOSE("rp_close ttyRP%d: Skipping FIFO drain - TXFIFO Empty\n", ChP->portNumber);
    }

    spin_lock_bh(&ChP->slock);
    ChP->xmit_flag = 0;
    spin_unlock_bh(&ChP->slock);

    //Set hardware to offline state

    //turn off the rcv fifo otherwise a flood of data to a previously opened then
    //closed port will render the read useless due to some unknown condition .. investigation needed!
    uDisRxFIFO(ChP);

    uDisTransmit(ChP);
    DisableInterrupts(ChP);
    SetFlowControl(ChP, FLOW_CONTROL_NONE);
    uSetXon(ChP);
    uFlushTxFifo(ChP);
    uFlushRxFifo(ChP);
    uClrRts(ChP);
    if (C_HUPCL(tty)) uClrDtr(ChP);

    if (TTY_DRIVER_FLUSH_BUFFER_EXISTS(tty))
        TTY_DRIVER_FLUSH_BUFFER(tty);

#if LINUX_VERSION_CODE < VERSION_CODE(2,6,26)
    if (tty->ldisc.flush_buffer)
        tty->ldisc.flush_buffer(tty);
#else           // implies kernel >= 2,6,26
    if (TTY_LDISC_FLUSH_BUFFER_EXISTS(tty))
        TTY_LDISC_FLUSH_BUFFER(tty);
#endif          // kernel <2,6,26

    if (ChP->blocked_open) {
        if (ChP->close_delay) {
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(msecs_to_jiffies(ChP->close_delay));
        }
        wake_up_interruptible(&ChP->open_wait);
    } else {
        if (ChP->xmit_buf) {
            free_page((unsigned long)ChP->xmit_buf);
            ChP->xmit_buf = 0;
        }
    }

    ChP->flags &= ~(ASYNC_INITIALIZED | ASYNC_CLOSING | ASYNC_NORMAL_ACTIVE);
    tty->closing = 0;
    wake_up_interruptible(&ChP->close_wait);

    atomic_dec(&num_ports_open);
    DOPEN("rp_close ttyRP%d - remaining numPortsOpen: %d \n", ChP->portNumber, atomic_read(&num_ports_open));
}

// write sycall handle - TTY write routine, called when user app
// writes to the device.  A per port write semaphore is used to
// protect from another process writing to this port at the same time.
// This other process could be running on the other CPU or get control
// of the CPU if the copy_from_user() blocks due to a page fault
// (swapped out).  Spinlocks protect the info xmit members.

#if LINUX_VERSION_CODE < VERSION_CODE(2,6,10)
static int rp_write(struct tty_struct *tty, int from_user,
                    const unsigned char *buf, int count)
#else           //implies kernel >= 2,6,10
static int rp_write(struct tty_struct *tty, const unsigned char *buf, int count)
#endif          //kernel < 2.6.10
{
    CHANNEL_T *ChP = (CHANNEL_T *)tty->driver_data;
    int numToXmit = 0, ret = 0;
    unsigned int *dPtr;
    unsigned char *bPtr;
    int i = 0, remainder = 0, steps = 0;

    //printk(KERN_INFO "rp_write: Begin call with count = %d\n", count);

    if (count == 0)
        return 0;

    if (count < 0) {
        printk(KERN_INFO "rocket2: write count == %d\n", count);
        return 0;
    }

    if(rocket_paranoia_check(ChP, "rp_write")) {
        printk(KERN_INFO "rocket2: FAILED Paranoia Check!\n");
        return 0;
    }

// copy_from_user cannot be called inside a spinlock
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,10)
    if (from_user) {
        count = MIN(count, PAGE_SIZE);
        if (down_interruptible(&ChP->tmp_buf_sem)) {
            printk(KERN_INFO "rp_write: down_interruptible FAILED!\n");
            printk(KERN_INFO "rp_write: Exiting call with -ERESTARTSYS\n");
            return -ERESTARTSYS;
        }
        if (buf && copy_from_user(ChP->tmp_buf, buf, count)) {
            printk(KERN_INFO "rp_write: Copy from_user FAILED!\n");
            up(&ChP->tmp_buf_sem);
            printk(KERN_INFO "rp_write: Exiting call with -EFAULT\n");
            return -EFAULT;
        } else if (!buf) {
            printk(KERN_INFO "rp_write: buffer empty .. exiting call.\n");
            return 0;
        }
        if(ChP->tmp_buf) {
            buf = ChP->tmp_buf;
        } else {
            printk(KERN_INFO "rp_write: temporary buffer empty .. exiting call.\n");
            return 0;
        }
    }
#endif          //kernel < 2,6,10

    spin_lock_bh(&ChP->slock);

    DWRITE( "rp_write on ttyRP%d - %d chars xmit_cnt: %d  xmit_fifo_room: %d \n", ChP->portNumber, count, ChP->xmit_cnt, ChP->xmit_fifo_room);

    if (!tty->stopped && !tty->hw_stopped && ChP->xmit_fifo_room < count) {
        ChP->xmit_fifo_room = FIFO_SIZE - uGetTxCnt(ChP);
        //printk(KERN_INFO "rp_write: xmit_fifo_room = %d\n", ChP->xmit_fifo_room);
    }

    // If the write queue for the port is empty, and there is FIFO space, stuff bytes
    // into FIFO. Use the write queue for temp storage.

    if (!tty->stopped && !tty->hw_stopped && ChP->xmit_cnt == 0 && ChP->xmit_fifo_room > 0) {
        numToXmit = MIN(count, ChP->xmit_fifo_room);
        dPtr = (unsigned int *)(buf);
        steps = numToXmit / 4;
        remainder = numToXmit % 4;

        if (steps) {
            DWRITE( "rp_write ttyRP%d - writing %d dwords \n", ChP->portNumber, steps);
            for (i = 0; i < steps; i++) {
                uXmitDword(ChP, *dPtr++);
            }
        }

        if (remainder) {
            bPtr = (unsigned char *)dPtr;
            for (i = 0; i < remainder; i++) {
                uXmitByte(ChP, *bPtr++);
            }
        }

        ret += numToXmit;
        buf += numToXmit;
        count -= numToXmit;
        ChP->xmit_fifo_room -= numToXmit;
    }

    // If count is zero, we wrote it all and are done
    if (!count) goto end;

    //  Write remaining data into the port's xmit_buf
    while (1) {
        if (ChP->rptty == 0) goto end;

        numToXmit = MIN(count, MIN(XMIT_BUF_SIZE - ChP->xmit_cnt - 1, XMIT_BUF_SIZE - ChP->xmit_head));
        if (numToXmit <= 0) break;

        // do not call copy_from_user() while inside a spinlock because
        // that function can sleep and will cause kernel oops. Also it
        // is not needed here since we alread transfered as much as we
        // can handle in one shot to buf. if there is any remaining
        // in buf then store it and move on.

        memcpy(ChP->xmit_buf + ChP->xmit_head, buf, numToXmit);

        ChP->xmit_head = (ChP->xmit_head + numToXmit) & (XMIT_BUF_SIZE - 1);
        ChP->xmit_cnt += numToXmit;
        buf += numToXmit;
        count -= numToXmit;
        ret += numToXmit;
    }

    if ((ret > 0) && !tty->stopped && !tty->hw_stopped) {
        ChP->xmit_flag = 1;
    }

end:

    spin_unlock_bh(&ChP->slock);

#if LINUX_VERSION_CODE < VERSION_CODE(2,6,10)
    if (from_user) up(&ChP->tmp_buf_sem);
#endif          //kernel < 2,6,10

    if (ChP->xmit_cnt < WAKEUP_CHARS) {
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,10)
        if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) && tty->ldisc.write_wakeup) {
            (tty->ldisc.write_wakeup) (tty);
        }
        wake_up_interruptible(&tty->write_wait);
#else
        tty_wakeup(tty);
#endif
    }

    //printk(KERN_INFO "rp_write: End call\n");

    return ret;
}


// syscall handler - write char routine.  Write buffer and counters
// protected by spinlocks

#if LINUX_VERSION_CODE < VERSION_CODE(2,6,26)
static void rp_put_char(struct tty_struct *tty, unsigned char ch)
#else           //implies kernel >= 2,6,26
static int rp_put_char(struct tty_struct *tty, unsigned char ch)
#endif          //kernel < 2,6,26
{
    int ret = 0;

    CHANNEL_T *ChP = (CHANNEL_T *)tty->driver_data;

    //printk(KERN_INFO "----------DEBUG----------- Entering rp_put_char()\n");

    if (rocket_paranoia_check(ChP, "rp_put_char")) {
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,26)
        return;
#else
        return ret;
#endif
    }
    DWRITE("rp_put_char ttyRP%d - char: %c (0x%x) \n", ChP->portNumber, ch, ch);


    spin_lock_bh(&ChP->slock);

    // If the write queue for the port is empty, and there is FIFO
    // space, stuff byte into FIFO.  Use the write queue for temp
    // storage.

    ChP->xmit_fifo_room = FIFO_SIZE - uGetTxCnt(ChP);
    if (tty->stopped || tty->hw_stopped || ChP->xmit_fifo_room == 0 || ChP->xmit_cnt != 0) {
        ChP->xmit_buf[ChP->xmit_head++] = ch;
        ChP->xmit_head &= XMIT_BUF_SIZE - 1;
        ChP->xmit_cnt++;
        ChP->xmit_flag = 1;
    } else {
        uXmitByte(ChP, ch);
        ChP->xmit_fifo_room--;
        ret++;
    }

    spin_unlock_bh(&ChP->slock);

#if LINUX_VERSION_CODE > VERSION_CODE(2,6,25)
    return ret;
#endif
}


// Return the number of characters that can be sent.  We estimate only
// using the in-memory transmit buffer only, and ignore the potential
// space in the transmit FIFO.

static int rp_write_room(struct tty_struct *tty)
{
    CHANNEL_T *ChP = (CHANNEL_T *)tty->driver_data;
    int ret;

    if (rocket_paranoia_check(ChP, "rp_write_room"))
        return 0;

    ret = XMIT_BUF_SIZE - ChP->xmit_cnt - 1;
    if (ret < 0)
        ret = 0;
    DWRITE("rp_write_room ttyRP%d returns %d \n", ChP->portNumber, ret);
    return ret;
}


// Return the number of characters in the buffer.  Again, this only
// counts those characters in the in-memory transmit buffer.

static int rp_chars_in_buffer(struct tty_struct *tty)
{
    CHANNEL_T *ChP = (CHANNEL_T *)tty->driver_data;

    if (rocket_paranoia_check(ChP, "rp_chars_in_buffer"))
        return 0;

    //DREAD("rp_chars_in_buffer ttyRP%d returns %d \n", ChP->portNumber, ChP->xmit_cnt);
    return ChP->xmit_cnt;
}


// Flushes the TX fifo for a port, deletes data in the xmit_buf stored
// in the r_port struct for the port.

static void rp_flush_buffer(struct tty_struct *tty)
{
    CHANNEL_T *ChP = (CHANNEL_T *)tty->driver_data;


    if (rocket_paranoia_check(ChP, "rp_flush_buffer"))
        return;

    DIO("rp_flush_buffer ttyRP%d \n", ChP->portNumber);

    ChP->xmit_cnt = ChP->xmit_head = ChP->xmit_tail = 0;

#if LINUX_VERSION_CODE < VERSION_CODE(2,6,10)
    if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) && tty->ldisc.write_wakeup) {
        (tty->ldisc.write_wakeup) (tty);
    }
    wake_up_interruptible(&tty->write_wait);
#else
    tty_wakeup(tty);
#endif

    uFlushTxFifo(ChP);
}

// IOCTL call handler
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,39)
static int rp_ioctl(struct tty_struct *tty, struct file *file, unsigned int cmd, unsigned long arg)
#else
static int rp_ioctl(struct tty_struct *tty, unsigned int cmd, unsigned long arg)
#endif
{
    CHANNEL_T *ChP = (CHANNEL_T *)tty->driver_data;

    //DIO( "rp_ioctl on ttyRP%d - cmd: %d \n", ChP->portNumber, cmd);
    if (cmd != RCKP_GET_PORTS && rocket_paranoia_check(ChP, "rp_ioctl"))
        return -ENXIO;

    switch (cmd) {
    case TIOCMGET:
        return get_modem_info(ChP, (unsigned int *) arg);
    case TIOCMBIS:
    case TIOCMBIC:
    case TIOCMSET:
        return set_modem_info(ChP, cmd, (unsigned int *) arg);

    case TIOCGSERIAL:
        return get_serial_info(ChP, (struct serial_struct*)arg);
    case TIOCSSERIAL:
        return set_serial_info(ChP, (struct serial_struct*)arg);

    case RCKP_GET_STRUCT:
        if (copy_to_user((void *) arg, ChP, sizeof (*ChP)))
            return -EFAULT;
        return 0;

    case RCKP_GET_CONFIG:
        return get_config(ChP, (struct rocket_config *) arg);
    case RCKP_SET_CONFIG:
        return set_config(ChP, (struct rocket_config *) arg);
    case RCKP_GET_PORTS:
        return get_ports(ChP, (struct rocket_ports *) arg);
    case RCKP_GET_VERSION:
        return get_version(ChP, (struct rocket_version *) arg);

#ifdef MODULE
        //  Resets the use count, incase the module crashes.  Avoids having to reboot
    case RP2_RESET_USE_COUNT:
        printk("RPInfinity:  Resetting USE Count \n");
        cleanup_module();
        break;

#endif //  MODULE

    default:
        //DIO( "rp_ioctl on ttyRP%d - Invalid IOCTL received: %d \n", ChP->portNumber, cmd);
        return -ENOIOCTLCMD;
    }
    return 0;
}

static void rp_throttle(struct tty_struct *tty)
{
    CHANNEL_T *ChP = (CHANNEL_T *)tty->driver_data;

    if (rocket_paranoia_check(ChP, "rp_throttle"))
        return;
    DIO("rp_throttle ttyRP%d \n", ChP->portNumber);

    if (I_IXOFF(tty))
        rp_send_xchar(tty, STOP_CHAR(tty));
    uClrRts(ChP);
}

static void rp_unthrottle(struct tty_struct *tty)
{
    CHANNEL_T *ChP = (CHANNEL_T *)tty->driver_data;

    if (rocket_paranoia_check(ChP, "rp_unthrottle"))
        return;
    DIO("rp_unthrottle ttyRP%d \n", ChP->portNumber);

    if (I_IXOFF(tty))
        rp_send_xchar(tty, START_CHAR(tty));

    uSetRts(ChP);
}

// Configures the serial port based on the TTY layer settings

#if LINUX_VERSION_CODE < VERSION_CODE(2,6,20)
static void rp_set_termios(struct tty_struct *tty, struct termios *old_termios)
#else           //implies kernel >= 2,6,20
static void rp_set_termios(struct tty_struct *tty, struct ktermios *old_termios)
#endif          //kernel < 2,6,20
{
    CHANNEL_T *ChP = (CHANNEL_T *)tty->driver_data;

    if (rocket_paranoia_check(ChP, "rp_set_termios"))
        return;
    DIO("rp_set_termios ttyRP%d \n", ChP->portNumber);

    configure_r_port(ChP, old_termios);

    // Handle transition to B0 status
    if ((old_termios->c_cflag & CBAUD) && !(tty->termios->c_cflag & CBAUD)) {
        uClrDtr(ChP);
        uClrRts(ChP);
    }

    // Handle transition away from B0 status
    if (!(old_termios->c_cflag & CBAUD) && (tty->termios->c_cflag & CBAUD)) {
        if (!tty->hw_stopped)
            uSetRts(ChP);
        uSetDtr(ChP);
    }

    if ((old_termios->c_cflag & CRTSCTS) && !(tty->termios->c_cflag & CRTSCTS)) {
        tty->hw_stopped = 0;
        rp_start(tty);
    }
}



// Stops the transmitter, if there is data in the TX FIFO.

static void rp_stop(struct tty_struct *tty)
{
    CHANNEL_T *ChP = (CHANNEL_T *)tty->driver_data;

    if (rocket_paranoia_check(ChP, "rp_stop"))
        return;
    DIO("rp_stop ttyRP%d \n", ChP->portNumber);
    if (uGetTxCnt(ChP))
        uDisTransmit(ChP);
}

// Starts the transmitter.

static void rp_start(struct tty_struct *tty)
{
    CHANNEL_T *ChP = (CHANNEL_T *)tty->driver_data;

    if (rocket_paranoia_check(ChP, "rp_start"))
        return;
    DIO("rp_start ttyRP%d \n", ChP->portNumber);
    uEnTransmit(ChP);
    ChP->xmit_flag = 1;
}


// rp_hangup() --- called by tty_hangup() when a hangup is signaled.

static void rp_hangup(struct tty_struct *tty)
{
    CHANNEL_T *ChP = (CHANNEL_T *)tty->driver_data;

    if (rocket_paranoia_check(ChP, "rp_hangup"))
        return;
    DIO("rp_hangup ttyRP%d \n", ChP->portNumber);

    rp_flush_buffer(tty);
    if (ChP->flags & ASYNC_CLOSING)
        return;

    if (ChP->count) {
        atomic_dec(&num_ports_open);
    }
    ChP->xmit_flag = 0;
    ChP->count = 0;
    ChP->flags &= ~ASYNC_NORMAL_ACTIVE;
    ChP->rptty = 0;

    uDisTransmit(ChP);
    DisableInterrupts(ChP);
    SetFlowControl(ChP, FLOW_CONTROL_NONE);
    uSetXon(ChP);
    ChP->flags &= ~ASYNC_INITIALIZED;

    wake_up_interruptible(&ChP->open_wait);
}


#if LINUX_VERSION_CODE < VERSION_CODE(2,6,27)
static void rp_break(struct tty_struct *tty, int break_state)
#else           // implies kernel >= 2,6,27
static int rp_break(struct tty_struct *tty, int break_state)
#endif          // kernel < 2,6,27
{
    CHANNEL_T *ChP = (CHANNEL_T *)tty->driver_data;

    if (rocket_paranoia_check(ChP, "rp_break")) {
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,27)
        return;
#else
        return 0;
#endif          // kernel <2,6,27
    }

    DIO("rp_break ttyRP%d \n", ChP->portNumber);

    if (break_state == -1)
        uSendBreak(ChP);
    else
        uClrBreak(ChP);

#if LINUX_VERSION_CODE < VERSION_CODE(2,6,27)
    return;
#else           //implies kernel >= 2,6,27
    return 1;
#endif          //kernel < 2,6,27
}


static void rp_send_xchar(struct tty_struct *tty, char ch)
{
    CHANNEL_T *ChP = (CHANNEL_T *)tty->driver_data;

    if (rocket_paranoia_check(ChP, "rp_send_xchar"))
        return;
    DIO("rp_send_xchar ttyRP%d \n", ChP->portNumber);

    if (uGetTxCnt(ChP))
        uXmitBytePriority(ChP, ch);
    else
        uXmitByte(ChP, ch);
}


// Wait until the transmitter is empty.  Uses ChP->cps (baud/bits) to
// calculate when the data should be flushed from the FIFO.  timeout
// == how long to wait before giving up.

static void rp_wait_until_sent(struct tty_struct *tty, int timeout)
{
    CHANNEL_T *ChP = (CHANNEL_T *)tty->driver_data;
    int txFifoCnt;
    int check_time; // jiffies required to send (txFifoCnt) characters
    int exit_time;  // total jiffies has elapsed
    unsigned long orig_jiffies;

    if (rocket_paranoia_check(ChP, "rp_wait_until_sent"))
        return;
    DWSENT("rp_wait_until_sent ttyRP%d - timeout (msec): %d txFifoCnt: %d cps: %d \n",
           ChP->portNumber, (timeout * 1000/ HZ), uGetTxCnt(ChP), ChP->cps);

    // Sanity check.  Max timeout is (FIFO_SIZE + 1)/30 CPS (300 baud)
    // in HZ, which is 17.1 (~18) seconds.  TTY layer often sends in
    // BIGNUMBER to wait forever, we need to catch that and make the
    // delay reasonable.

    if (timeout > 18 * HZ) {
        //DWSENT("RP2:  rp_wait_until_sent() ttyRP%d - invalid timeout value: %lu (msec) txFifoCnt: %d cps: %d \n",
        //        ChP->portNumber, timeout * 1000 / HZ, uGetTxCnt(ChP), ChP->cps);

        // Our usual calculation, time to transmit (txFifoCnt +1) bytes in jiffies
        timeout = ((((uGetTxCnt(ChP) + 1) * HZ) / ChP->cps));
        if (timeout <= 0)
            timeout = 1;
        if (timeout > 18 * HZ)
            timeout = 18 * HZ;
        //DWSENT("rp_wait_until_sent() ttyRP%d - new timeout value calculated: %lu (msec) \n",  ChP->portNumber, ((timeout * 1000) / HZ));
    }

    // Other drivers use 2 * calculated timeout value, so we will too.
    // Testing showed this results in the FIFO being drained properly,
    // if we are not being flowed off.

    timeout *= 2;
    orig_jiffies = jiffies;
    while (1) {
        txFifoCnt = uGetTxCnt(ChP);

        if (txFifoCnt > FIFO_SIZE) {
            DWSENT("rp_wait_until_sent ttyRP%d - TX FIFO Count > fifo size: %d \n", ChP->portNumber, txFifoCnt);
            break;
        }

        if (!txFifoCnt) {
            if (uXmitShiftRegEmpty(ChP))
                break;
            check_time = (HZ / ChP->cps) / 5;   // One byte
        } else {
            check_time = HZ * txFifoCnt / ChP->cps;
        }

        if (timeout) {
            // work-around for broken tty-layer code that passes 64-bit
            // value resulting in our seeing -1 for a timeout instead of
            // max signed value.
            if (timeout > 0)
                exit_time = orig_jiffies + timeout - jiffies;
            else
                exit_time = ((unsigned)~0)>>1;
            if (exit_time <= 0)
                break;
            if (exit_time < check_time)
                check_time = exit_time;
        }
        if (check_time == 0)
            check_time = 1;
        //DIO("rp_wait_until_sent ttyRP%d entering delay - txFifoCnt: %d jiffies: %lu check_time: %d \n",  ChP->portNumber, txFifoCnt, jiffies, check_time);
        current->state = TASK_INTERRUPTIBLE;
        schedule_timeout(check_time);
        if (signal_pending(current))
            break;
    }
    DWSENT("rp_wait_until_sent ttyRP%d done - txFifoCnt: %d \n",  ChP->portNumber, txFifoCnt);
}


// Returns the state of the serial modem control lines.  These next 2
// functions are the way kernel versions > 2.5 handle modem control
// lines rather than IOCTLs.

#if LINUX_VERSION_CODE < VERSION_CODE(2,6,39)
static int rp_tiocmget(struct tty_struct *tty, struct file *file)
#else
static int rp_tiocmget(struct tty_struct *tty)
#endif
{
    CHANNEL_T *ChP = (CHANNEL_T *)tty->driver_data;
    return uGetMsr(ChP);
}

// Sets the modem control lines

#if LINUX_VERSION_CODE < VERSION_CODE(2,6,39)
static int rp_tiocmset(struct tty_struct *tty, struct file *file, unsigned int set, unsigned int clear)
#else
static int rp_tiocmset(struct tty_struct *tty, unsigned int set, unsigned int clear)
#endif
{
    CHANNEL_T *ChP = (CHANNEL_T *)tty->driver_data;

    if (set & TIOCM_RTS)
        uSetRts(ChP);
    if (set & TIOCM_DTR)
        uSetDtr(ChP);
    if (clear & TIOCM_RTS)
        uClrRts(ChP);
    if (clear & TIOCM_DTR)
        uClrDtr(ChP);
    return 0;
}



// Blocks, if O_NONBLOCK is clear, waiting for the CD (perhaps other
// things also.)  Sets ASYNC_NORMAL_ACTIVE if return value is 0.
// Returns: 0 if OK, retval else.

static int block_til_ready(struct tty_struct *tty, struct file *filp, CHANNEL_T *ChP)
{
    DECLARE_WAITQUEUE(wait, current);
    int do_clocal = 0, retval = 0;

    // If the device is in the middle of being closed, then block until
    // it's done, and then try again.

    if (tty_hung_up_p(filp))
        return ((ChP->flags & ASYNC_HUP_NOTIFY) ? -EAGAIN : -ERESTARTSYS);
    if (ChP->flags & ASYNC_CLOSING) {
        wait_event_interruptible(ChP->close_wait,(!(ChP->flags&ASYNC_CLOSING)));
        return ((ChP->flags & ASYNC_HUP_NOTIFY) ? -EAGAIN : -ERESTARTSYS);
    }

    // If non-blocking mode is set, or the port is not enabled, then
    // make the check up front and then exit.

    if ((filp->f_flags & O_NONBLOCK) || (tty->flags & (1 << TTY_IO_ERROR))) {
        ChP->flags |= ASYNC_NORMAL_ACTIVE;
        return 0;
    }
    if (tty->termios->c_cflag & CLOCAL)
        do_clocal = 1;

    // Block waiting for the carrier detect and the line to become free.

    add_wait_queue(&ChP->open_wait, &wait);
    DOPEN("block_til_ready before block: ttyRP%d, count = %d\n", ChP->portNumber, ChP->count);
    ChP->blocked_open++;

    while (1) {
        if (tty->termios->c_cflag & CBAUD) {
            uSetDtr(ChP);
            uSetRts(ChP);
        }
        set_current_state(TASK_INTERRUPTIBLE);
        if (tty_hung_up_p(filp) || !(ChP->flags & ASYNC_INITIALIZED)) {
            if (ChP->flags & ASYNC_HUP_NOTIFY)
                retval = -EAGAIN;
            else
                retval = -ERESTARTSYS;
            break;
        }
        if (!(ChP->flags & ASYNC_CLOSING) &&
                (do_clocal || (uCdActive(ChP))))
            break;
        if (signal_pending(current)) {
            retval = -ERESTARTSYS;
            break;
        }
        DOPEN("block_til_ready blocking: ttyR%d, count = %d, flags=0x%0x\n", ChP->portNumber, ChP->count, ChP->flags);
        schedule();     // Don't hold spinlock here, will hang PC
    }

    current->state = TASK_RUNNING;
    remove_wait_queue(&ChP->open_wait, &wait);
    ChP->blocked_open--;

    DOPEN("block_til_ready after blocking: ttyR%d, count = %d\n", ChP->portNumber, ChP->count);
    if (retval)
        return retval;
    ChP->flags |= ASYNC_NORMAL_ACTIVE;

    return 0;
}


// Routines used by rp_ioctl

// get/set serial info ioctls used by 'setserial' and similar programs

static int get_serial_info(CHANNEL_T *ChP, struct serial_struct * retinfo)
{
    struct serial_struct tmp;
    if (!retinfo)
        return -EFAULT;
    memset(&tmp, 0, sizeof(tmp));

    tmp.port               = ChP->channelNumber;
    tmp.line               = ChP->portNumber;
    tmp.irq                = 0;
    tmp.flags              = ChP->flags;
    tmp.baud_base          = ChP->baud_base;
    tmp.custom_divisor     = ChP->custom_divisor;
    tmp.close_delay        = ChP->close_delay / 10;
    tmp.closing_wait       = ChP->closing_wait == ASYNC_CLOSING_WAIT_NONE ? ASYNC_CLOSING_WAIT_NONE : ChP->closing_wait / 10;
    tmp.iomem_base         = (void*)ChP->pBaseRegister;

    if (copy_to_user(retinfo, &tmp, sizeof *retinfo))
        return -EFAULT;
    return 0;
}

static int set_serial_info(CHANNEL_T *ChP, struct serial_struct *usr_serial)
{
    struct serial_struct new_serial;
    if (copy_from_user(&new_serial,usr_serial,sizeof(new_serial)))
        return -EFAULT;

    // convert times from 1/100 second to milliseconds;
    new_serial.close_delay *= 10;
    if (new_serial.closing_wait != ASYNC_CLOSING_WAIT_NONE)
        new_serial.closing_wait *= 10;

#ifdef CAP_SYS_ADMIN
    if (!capable(CAP_SYS_ADMIN))
#else
    if (!suser())
#endif
    {
        if (new_serial.close_delay != ChP->close_delay ||
                (new_serial.flags & ROCKET_ALLFLAGS & ~ROCKET_USR_MASK) != (ChP->flags & ROCKET_ALLFLAGS & ~ROCKET_USR_MASK))
            return -EPERM;
        ChP->flags = (ChP->flags & ~ROCKET_USR_MASK) | (new_serial.flags & ROCKET_USR_MASK);
    } else {
        ChP->flags = (ChP->flags & ~ROCKET_ALLFLAGS) | (new_serial.flags & ROCKET_ALLFLAGS);
        ChP->close_delay = new_serial.close_delay;
        ChP->closing_wait = new_serial.closing_wait;
    }
    ChP->baud_base = new_serial.baud_base;
    ChP->custom_divisor = new_serial.custom_divisor;

    return 0;
}


// Gets the MSR values for the port, returns it to userspace through the value ptr.

static int get_modem_info(CHANNEL_T *ChP, unsigned int *value)
{
    unsigned int result;
    if (copy_to_user(value, &result, sizeof (int)))
        return -EFAULT;
    return 0;
}

// Sets the modem control lines, based on the cmd (the IOCTL) passed in and the user arg (the TTY command).

static int set_modem_info(CHANNEL_T *ChP, unsigned int cmd, unsigned int *value)
{
    unsigned int arg;

    if (copy_from_user(&arg, value, sizeof (int)))
        return -EFAULT;

    switch (cmd) {
        // Turns on the selected line
    case TIOCMBIS:
        if (arg & TIOCM_RTS)
            uSetRts(ChP);
        if (arg & TIOCM_DTR)
            uSetDtr(ChP);
        break;

        // Turns off the selected lines
    case TIOCMBIC:
        if (arg & TIOCM_RTS)
            uClrRts(ChP);
        if (arg & TIOCM_DTR)
            uClrDtr(ChP);
        break;

        // this one turns off RTS and DTR, then turns on either if requested
    case TIOCMSET:
        uClrRts(ChP);
        uClrDtr(ChP);
        if (arg & TIOCM_RTS)
            uSetRts(ChP);
        if (arg & TIOCM_DTR)
            uSetDtr(ChP);
        break;

    default:
        return -EINVAL;
    }
    return 0;
}



// Sanity check performed before performing TTY read/write/etc.
// Returns: 1 if OK, 0 else

static int rocket_paranoia_check(CHANNEL_T *ChP, const char *routine)
{
    if (!ChP)
        return 1;
    if (ChP->magic != RPORT_MAGIC) {
        printk(KERN_INFO "Warning: bad magic number for rocketport struct in %s\n",  routine);
        return 1;
    }
    return 0;
}

// Set flow control (FLOW_CONTROL_xxx) for the port

static void SetFlowControl(CHANNEL_T *ChP, int flowControl)
{
    switch (flowControl) {
    case FLOW_CONTROL_NONE:
        DOPEN("SetFlowControl to NONE  ttyRP%d \n", ChP->portNumber);
        uDisRtsFlowCtrl(ChP);
        uDisCtsFlowCtrl(ChP);
        uDisDtrFlowCtrl(ChP);
        uDisDsrFlowCtrl(ChP);
        uDisSoftFlowCtrl(ChP);
        break;

    case FLOW_CONTROL_RTSCTS:
        DOPEN("SetFlowControl to RTS/CTS  ttyRP%d \n", ChP->portNumber);
        uEnRtsFlowCtrl(ChP);
        uEnCtsFlowCtrl(ChP);
        uDisDtrFlowCtrl(ChP);
        uDisDsrFlowCtrl(ChP);
        uDisSoftFlowCtrl(ChP);
        break;

    case FLOW_CONTROL_DTRDSR:
        DOPEN("SetFlowControl to DTR/DSR  ttyRP%d \n", ChP->portNumber);
        uDisRtsFlowCtrl(ChP);
        uDisCtsFlowCtrl(ChP);
        uEnDtrFlowCtrl(ChP);
        uEnDsrFlowCtrl(ChP);
        uDisSoftFlowCtrl(ChP);
        break;

    case FLOW_CONTROL_XONXOFF:
        DOPEN("SetFlowControl to XON/XOFF  ttyRP%d \n", ChP->portNumber);
        uDisRtsFlowCtrl(ChP);
        uDisCtsFlowCtrl(ChP);
        uDisDtrFlowCtrl(ChP);
        uDisDsrFlowCtrl(ChP);
        uEnSoftFlowCtrl(ChP);
        break;
    }
}

// Set the baud divisor for the requested baud rate (or at least as
// close as we can get).

static int SetBaudRate(CHANNEL_T *ChP, int baudRate)
{
    int baudDivisor = AiopBaudDivisor(baudRate);
    if (baudDivisor < 0)
        baudDivisor = 0;
    DIO("SetBaudRate ttyRP%d baud: %d divisor: %d\n", ChP->portNumber, baudRate, baudDivisor);
    uSetBaudDivisor(ChP,  baudDivisor);
    return 0;
}

static int GetNumberOfModels(void)
{
    return (sizeof(modelInfo) / sizeof(modelInfo_t));
}

// Gets the modelInfo struct corresponding to the PCI model ID

static modelInfo_t *GetModelInfo(unsigned long model)
{
    int i;

    for (i = 0; i < GetNumberOfModels(); i++) {
        if (modelInfo[i].model ==  model)
            return &modelInfo[i];
    }
    printk("DEBUG: modelInfo for model::%ld not found!\n", model);//BKA struct array malformed?
    return NULL;
}

// Retrieves the channel structure for a give line number.

static CHANNEL_T *GetChannelStruct(int line)
{
    int i;

    for (i = 0; i < numberof_boards; i++) {
        if ((line >= uController[i].rocketModel.startingPortNumber) &&
                (line < uController[i].rocketModel.startingPortNumber + uController[i].numPorts))
            return &uController[i].channel[line - uController[i].rocketModel.startingPortNumber];
    }
    return NULL;
}


// Enables/disables the standard interrupts used by the driver.
// These are:
//  -  TX FIFO Empty
//  -  RX Trigger
//  -  Special Receive
//  -  Modem Control Change
//
// The channel is then unmasked/masked in the Channel Interrupt Mask.

static void EnableInterrupts(CHANNEL_T *ChP)
{
    uEnChannelInterrupt(ChP);
    uSetRxTrigger(ChP, URCV_FIFO_TRIG_DISABLED);
    uEnSpecialReceiveInt(ChP);
    uEnModemChangeInt(ChP);
}

static void DisableInterrupts(CHANNEL_T *ChP)
{
    uDisXmitFifoEmptyInt(ChP);
    uSetRxTrigger(ChP,  URCV_FIFO_TRIG_DISABLED);
    uDisSpecialReceiveInt(ChP);
    uDisModemChangeInt(ChP);
    uDisChannelInterrupt(ChP);
}

// Gets the MSR values for the port as needed by the TTY layer.

static unsigned long uGetMsr(CHANNEL_T * ChP)
{
    unsigned long result = 0;

    if (uRtsActive(ChP))
        result |= TIOCM_RTS;
    if (uDtrActive(ChP))
        result |= TIOCM_DTR;
    if (uCdActive(ChP))
        result |= TIOCM_CAR;
    if (uRiActive(ChP))
        result |= TIOCM_RNG;
    if (uDsrActive(ChP))
        result |= TIOCM_DSR;
    if (uCtsActive(ChP))
        result |= TIOCM_CTS;

    //DIO("uGetMsr() ttyRP%d - retrieved msr: 0x%x \n",  ChP->portNumber, result);
    return result;
}


